/*  Fanny Xu, CS812, Phase 3
 *
 *  encodeExpr.c - code generation for expressions
 *
 */

#include "encode.h"
#include "message.h"
#include "analyze.h"

/*  70% level  */

/* This function does the encoding for the binary '+' operator.
 */
void encodeAdd(BinopTree b)
{
  TypeTag myType;
  
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeAdd\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeAdd\n");

  b->expn.reg = setReg();
  
  myType = typeQuery(b->expn.type);
  if( myType == TYERROR ) return; /* should never happen*/
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$4, %%esp");
    emit("\tflds\t(%%esp)");
    emit("\tfaddp");
    emit("\tfstps\t(%%esp)");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfldl\t(%%esp)");
    emit("\tfaddp");
    emit("\tfstpl\t(%%esp)");
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$12, %%esp");
    emit("\tfldt\t(%%esp)");
    emit("\tfaddp");
    emit("\tfstpt\t(%%esp)");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\taddl\t%%edx,%%eax");
    emit("\tpushl\t%%eax");
    return;
  }

  /* rel start */
  case TYPOINTER:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\taddl\t%%edx,%%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  /* rel end */

  default:
    bugT((Tree)b, "Invalid tree type in encodeAdd() %d", myType);
 }
  return;
}

/* This function does the encoding for the binary '-' operator.
 */
void encodeSub(BinopTree b)
{
  TypeTag myType;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "Null pointer passed to encodeSub\n");
  if( !b->left || !b->right )
    bugT( (Tree)b, "Null child pointer passed to encodeSub\n");

  b->expn.reg = setReg();
  myType = typeQuery(b->expn.type);
  if( myType == TYERROR) return;
  switch( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$4, %%esp");
    emit("\tflds\t(%%esp)");
    emit("\tfsubp");
    emit("\tfstps\t(%%esp)");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfldl\t(%%esp)");
    emit("\tfsubp");
    emit("\tfstpl\t(%%esp)");
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$12, %%esp");
    emit("\tfldt\t(%%esp)");
    emit("\tfsubp");
    emit("\tfstpt\t(%%esp)");
    return;
  }
  
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\tsubl\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  } 

  case TYPOINTER:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\tsubl\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
 
  default:
    bugT((Tree)b, "Invalid tree type in encodeSub()");
 }
  return;
}

/* This function does the encoding for the binary operator '/'.
 */
void encodeDiv(BinopTree b)
{
  TypeTag myType;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "Null pointer passed to encodeDiv\n");
  if( !b->left || !b->right )
    bugT( (Tree)b, "Null child pointer passed to encodeDiv\n");

  b->expn.reg = setReg();
  myType = typeQuery(b->expn.type);
  if( myType == TYERROR) return;
  switch( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$4, %%esp");
    emit("\tflds\t(%%esp)");
    emit("\tfdivp");
    emit("\tfstps\t(%%esp)");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfldl\t(%%esp)");
    emit("\tfdivp");
    emit("\tfstpl\t(%%esp)");
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$12, %%esp");
    emit("\tfldt\t(%%esp)");
    emit("\tfdivp");
    emit("\tfstpt\t(%%esp)");
    return;
  }
  
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYSIGNEDCHAR:
  { 
    emit("\tpopl\t%%ecx");
    emit("\tpopl\t%%eax");
    emit("\tcltd");
    emit("\tidivl\t%%ecx,%%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  {
    emit("\tpopl\t%%ecx");
    emit("\tpopl\t%%eax");
    emit("\tmovl\t$0, %%edx");
    emit("\tdivl\t%%ecx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  } 
  default: 
    bugT((Tree)b, "Invalid tree type in encodeDiv()");
 }

  return;
}

/* This function does the encoding for the binary operator '*'.
 */
void encodeMult(BinopTree b)
{
  TypeTag myType;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "Null pointer passed to encodeSub\n");
  if( !b->left || !b->right )
    bugT( (Tree)b, "Null child pointer passed to encodeSub\n");

  b->expn.reg = setReg();
  myType = typeQuery(b->expn.type);
  if( myType == TYERROR) return;
  switch( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$4, %%esp");
    emit("\tflds\t(%%esp)");
    emit("\tfmulp");
    emit("\tfstps\t(%%esp)");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfldl\t(%%esp)");
    emit("\tfmulp");
    emit("\tfstpl\t(%%esp)");
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$12, %%esp");
    emit("\tfldt\t(%%esp)");
    emit("\tfmulp");
    emit("\tfstpt\t(%%esp)");
    return;
  }
  
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\timull\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  } 
  default:
    bugT( (Tree)b, "Invalid tree type in encodeMult()");
 }

  return;
}

/* This function does the encoding for the binary '%' operator.
 */
void encodeMod(BinopTree b)
{
  TypeTag myType;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "Null pointer passed to encodeSub\n");
  if( !b->left || !b->right )
    bugT( (Tree)b, "Null child pointer passed to encodeSub\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(b->expn.type);
  if( myType == TYERROR) return;
  switch( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%ecx");
    emit("\tpopl\t%%eax");
    emit("\tcltd");
    emit("\tidivl\t%%ecx, %%eax");
    emit("\tpushl\t%%edx");
    return;
  }

  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  {
    emit("\tpopl\t%%ecx");
    emit("\tpopl\t%%eax");
    emit("\tmovl\t$0, %%edx");
    emit("\tdivl\t%%ecx, %%eax");
    emit("\tpushl\t%%edx");
    return;
  }
  default:
    break; 
 }
}

/* Dummy function for encoding '*='. It is never called.
 */
void encodeMultAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeMultAssign: You are not supposed to get here!");
  return;
}

/* Dummy function for encoding '/='. It is never called.
 */
void encodeDivAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeDivAssign: You are not supposed to get here!");
  return;
}

/* Dummy function for encoding '%='. It is never called.
 */
void encodeModAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeModAssign: You are not supposed to get here!");
  return;
}

/* Dummy function for encoding '+='. It is never called.
 */
void encodeAddAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeAddAssign: You are not supposed to get here!");
  return;
}

/* Dummy function for encoding '-='. It is never called.
 */
void encodeSubAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeSubAssign: You are not supposed to get here!");
  return;
}

/* This function does the encoding for the unary '-' operator.
 */
void encodeUsub(UnopTree u)
{
  TypeTag myType;

  /* error checkings */
  if( !u )
    bugT( (Tree)u, "Null pointer passed to encodeUsub\n");
  if( !u->left )
    bugT( (Tree)u, "Null child pointer passed to encodeUsub\n");

  u->expn.reg = setReg();
  myType = typeQuery(u->expn.type);
  if( myType == TYERROR ) return;
  switch( myType ){
  case TYFLOAT:
  {
    emit("\tpopl\t%%eax");
    emit("\txorl\t$0x80000000, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }  
  case TYDOUBLE:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\txorl\t$0x80000000, %%eax");
    emit("\tpushl\t%%eax");
    emit("\tpushl\t%%edx");
    return;
  }    
  case TYLONGDOUBLE:
  {
    emit("\tpopl\t%%ecx");
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\txorl\t$0x8000, %%eax");
    emit("\tpushl\t%%eax");
    emit("\tpushl\t%%edx");
    emit("\tpushl\t%%ecx");
    return;
  }    
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%eax");
    emit("\tnegl\t%%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)u, "Invalid tree type in encodeUsub()\n");
 }
 return; 
}


/*  80% level  */

/* Dummy funtion to encode the unary operator '!'.
 * It is never called.
 */
void encodeLnot(UnopTree b)
{ 
  b = NULL; /* defeat warning */
  info("encodeLnot: You are not supposed to get here!");
  return; 
} 

/* This function does the encoding for the binary operator '<'.
 */
void encodeLt(BinopTree b)
{
  TypeTag myType;
 
  /* error checkings */ 
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeLt\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeLt\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t4(%%esp)");
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t8(%%esp)");
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t12(%%esp)");
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    if( myType == TYSIGNEDLONGINT || myType == TYSIGNEDSHORTINT ||
        myType == TYSIGNEDINT     || myType == TYSIGNEDCHAR )
      emit("\tsetl\t%%al");
    else
      emit("\tsetb\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type in encodeLt()");
 }
 return;
}

/* This function does the encoding for the binary operator '>'.
 */
void encodeGt(BinopTree b)
{
  TypeTag myType;
  
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeGt\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeGt\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\tflds\t4(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\tfldl\t8(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\tfldt\t12(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$69, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    if( myType == TYSIGNEDLONGINT || myType == TYSIGNEDSHORTINT ||
        myType == TYSIGNEDINT     || myType == TYSIGNEDCHAR )
      emit("\tsetg\t%%al");
    else
      emit("\tseta\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type in encodeGt()"); 
 }
 return;
}

/* This function does the encoding for the binary operator '<='.
 */
void encodeLte(BinopTree b)
{
  TypeTag myType;
  
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeLte\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeLte\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t4(%%esp)");
    emit("\tflds\t(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYDOUBLE:
  {
    emit("\tfldl\t8(%%esp)");
    emit("\tfldl\t(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t12(%%esp)");
    emit("\tfldt\t(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    if( myType == TYSIGNEDLONGINT || myType == TYSIGNEDSHORTINT ||
        myType == TYSIGNEDINT     || myType == TYSIGNEDCHAR )
      emit("\tsetle\t%%al");
    else
      emit("\tsetbe\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type in encodeLte()");
 }
 return;
}

/* This function does the encoding for the binary operator '>='.
 */
void encodeGte(BinopTree b)
{

  TypeTag myType;
 
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeGte\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeGte\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\tflds\t4(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\tfldl\t8(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\tfldt\t12(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    emit("\ttestb\t$5, %%ah");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    if( myType == TYSIGNEDLONGINT || myType == TYSIGNEDSHORTINT ||
        myType == TYSIGNEDINT     || myType == TYSIGNEDCHAR )
      emit("\tsetge\t%%al");
    else
      emit("\tsetae\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT( (Tree)b, "Invalid tree type in encodeGte()");
 }
 return;
}

/* This function does the encoding for the binary operator '=='.
 */
void encodeEql(BinopTree b)
{
  TypeTag myType;
  
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeEql\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeEql\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\tflds\t4(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    
    emit("\tandb\t$69, %%ah");
    emit("\tcmpb\t$64, %%ah");
    emit("\tsete\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\tfldl\t8(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");

    emit("\tandb\t$69, %%ah");
    emit("\tcmpb\t$64, %%ah");
    emit("\tsete\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\tfldt\t12(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");

    emit("\tandb\t$69, %%ah");
    emit("\tcmpb\t$64, %%ah");
    emit("\tsete\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDCHAR:
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    emit("\tsete\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type in encodeEql()");
 }
 return;
}

/* This function does the encoding for the binary operator '!='.
 */
void encodeNeql(BinopTree b)
{
  TypeTag myType;
  
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeNeql\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeNeql\n");

  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYFLOAT:
  {
    emit("\tflds\t(%%esp)");
    emit("\tflds\t4(%%esp)");
    emit("\taddl\t$8, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");
    
    emit("\tandb\t$69, %%ah");
    emit("\txorb\t$64, %%ah");
    emit("\tsetne\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYDOUBLE:
  {
    emit("\tfldl\t(%%esp)");
    emit("\tfldl\t8(%%esp)");
    emit("\taddl\t$16, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");

    emit("\tandb\t$69, %%ah");
    emit("\txorb\t$64, %%ah");
    emit("\tsetne\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");  
    return;
  }
  case TYLONGDOUBLE:
  {
    emit("\tfldt\t(%%esp)");
    emit("\tfldt\t12(%%esp)");
    emit("\taddl\t$24, %%esp");
    emit("\tfucompp");
    emit("\tfnstsw\t%%ax");

    emit("\tandb\t$69, %%ah");
    emit("\txorb\t$64, %%ah");
    emit("\tsetne\t%%al");

    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    emit("\tpopl\t%%eax");
    emit("\tpopl\t%%edx");
    emit("\tcmpl\t%%eax,%%edx");
    emit("\tsetne\t%%al");
    emit("\tmovzbl\t%%al, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type to encodeNeql()");
 }
 return;
}

/* This function does the encoding for the binary operator '&&'.
 */
void encodeLand(BinopTree b)
{
  TypeTag myType;
  Type newType;
  UnopTree uTree;
  int lab1, lab2;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeLand\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeLand\n");

  lab1 = genLabel();
  lab2 = genLabel();
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( typeQuery(((ExpTree)b->left)->type) == TYERROR ||
      typeQuery(((ExpTree)b->right)->type) == TYERROR )
  { return;
  }
 
  /* encodes the left operand first */
  switch ( myType ){
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  {
    newType = typeBuildBasic( myType, NO_QUAL );
    uTree = (UnopTree)newUnop( LNOT_OP, newType, b->left);
    encode( (Tree)analyzeLnot( uTree ));
    
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$0, %%eax");
    emit("\tjne\tL$%d", lab1);

    myType = typeQuery( ((ExpTree)b->right)->type);
    break;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    encode( (Tree)b->left );
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$0, %%eax");
    emit("\tje\tL$%d", lab1);/* if it's zero, && fails*/
 
    myType = typeQuery( ((ExpTree)b->right)->type);   
    break;
  }
  default:
    bugT((Tree)b, "Invalid left subtree type in encodeLand()");
 }

  /* encodes the right operand */
  switch( myType ){
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  {
    newType = typeBuildBasic( myType, NO_QUAL);
    uTree = (UnopTree)newUnop( LNOT_OP, newType, b->right);
    encode( (Tree)analyzeLnot( uTree));
    
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$0, %%eax");
    emit("\tjne\tL$%d", lab1);
    emit("\tpushl\t$%d", 1 );
    emit("\tjmp\tL$%d", lab2);
    emit("L$%d:", lab1 );
    emit("\tpushl\t$%d", 0);
    emit("L$%d:", lab2 );
    return;

  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
   encode( (Tree)b->right );
   emit("\tpopl\t%%eax");
   emit("\tcmpl\t$0, %%eax");
   emit("\tje\tL$%d", lab1);
   emit("\tpushl\t$%d", 1 );
   emit("\tjmp\tL$%d", lab2);
   emit("L$%d:", lab1 );
   emit("\tpushl\t$%d", 0);
   emit("L$%d:", lab2 );
   return;
  }
  default:
    bugT((Tree)b, "Invalid right subtree type in encodeLand()");
 }
 return;
}

/* This function does the encoding for the binary '||'.
 */ 
void encodeLor(BinopTree b)
{
  TypeTag myType;
  Type newType;
  UnopTree uTree;
  int lab1, lab2;

  /* error checkings */
  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeLor\n");
  if( !b->left || !b->right)
    bugT( (Tree)b, "NULL child pointer passed to encodeLor\n");

  lab1 = genLabel();
  lab2 = genLabel();
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( typeQuery(((ExpTree)b->left)->type ) == TYERROR ||
      typeQuery(((ExpTree)b->right)->type ) == TYERROR )
    return; 

  switch ( myType ){
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  {
    newType = typeBuildBasic( myType, NO_QUAL );
    uTree = (UnopTree)newUnop( LNOT_OP, newType, b->left);
    encode( (Tree)analyzeLnot( uTree ));
    
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$1, %%eax");
    emit("\tjne\tL$%d", lab1);
    
    myType = typeQuery( ((ExpTree)b->right)->type );
    break;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
    encode((Tree)b->left);
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$0, %%eax");
    emit("\tjne\tL$%d", lab1);
    
    myType = typeQuery( ((ExpTree)b->right)->type );
    break;

  }
  default:
    bugT( (Tree)b, "Invalid left subtree type in encodeLor()");
 }

  switch( myType ){
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  {
    newType = typeBuildBasic( myType, NO_QUAL );
    uTree = (UnopTree)newUnop( LNOT_OP, newType, b->right);
    encode( (Tree)analyzeLnot( uTree));
    
    emit("\tpopl\t%%eax");
    emit("\tcmpl\t$1, %%eax");
    emit("\tjne\tL$%d", lab1);
    emit("\tpushl\t$%d", 0 );
    emit("\tjmp\tL$%d", lab2);
    emit("L$%d:", lab1 );
    emit("\tpushl\t$%d", 1);
    emit("L$%d:", lab2 );
    return;
  }

  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  /* rel start */
  case TYPOINTER:
  /* rel end   */
  {
   encode((Tree)b->right );
   emit("\tpopl\t%%eax");
   emit("\tcmpl\t$0, %%eax");
   emit("\tjne\tL$%d", lab1);
   emit("\tpushl\t$%d", 0 );
   emit("\tjmp\tL$%d", lab2);
   emit("L$%d:", lab1 );
   emit("\tpushl\t$%d", 1);
   emit("L$%d:", lab2 );
   return;
  }
  default:
   bugT( (Tree)b, "Invalid right subtree type in encodeLor()");
 }

  return;
}

/*  90% level  */

/* This function does encoding for the unary operator '~'.
 */
void encodeComplmt(UnopTree u)
{
  TypeTag myType;

  if( !u )
    bugT( (Tree)u, "NULL pointer passed to encodeComplmt\n");
  if( !u->left)
    bugT( (Tree)u, "NULL child pointer passed to encodeComplmt\n");
  
  u->expn.reg = setReg();
  myType = typeQuery(((ExpTree)u->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%eax");
    emit("\tnotl\t%%eax"); 
    emit("\tpushl\t%%eax");
   return;
  }
  default:
    bugT( (Tree)u, "Invalid tree type in encodeComplmt()");
 }
  return;
}

/* This function does encoding for the binary operator '<<'.
 */
void encodeLshift(BinopTree b)
{
  TypeTag myType;

  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeLshift\n");
  if( !(b->left && b->right))
    bugT( (Tree)b, "NULL child pointer passed to encodeLshift\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tmovb\t(%%esp), %%cl");
    emit("\taddl\t$4, %%esp");
    emit("\tpopl\t%%eax");
    emit("\tsall\t%%cl, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "invalid tree type in encodeLshift()");
 }
  return;
}

/* This function does encoding for the binary operator '>>'.
 */
void encodeRshift(BinopTree b)
{
  TypeTag myType;

  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeRshift\n");
  if( !(b->left && b->right))
    bugT( (Tree)b, "NULL child pointer passed to encodeRshift\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tmovb\t(%%esp), %%cl");
    emit("\taddl\t$4, %%esp");
    emit("\tpopl\t%%eax");
    if( myType == TYSIGNEDLONGINT || myType == TYSIGNEDSHORTINT ||
        myType == TYSIGNEDINT     || myType == TYSIGNEDCHAR)
      emit("\tsarl\t%%cl, %%eax");
    else
      emit("\tshrl\t%%cl, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT( (Tree)b, "Invalid tree type in encodeRshift()");
 }
  return;
}

/* This function does encoding for the binary operator '&'.
 */
void encodeAndd(BinopTree b)
{
  TypeTag myType;

  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeAndd\n");
  if( !(b->left && b->right))
    bugT( (Tree)b, "NULL child pointer passed to encodeAndd\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return;
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\tandl\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT( (Tree)b, "Invalid tree type in encodeAndd()");
 }
 return;
}

/* This function does encoding for the binary operator '^'.
 */
void encodeXorr(BinopTree b)
{
  TypeTag myType;

  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeXorr\n");
  if( !(b->left && b->right))
    bugT( (Tree)b, "NULL child pointer passed to encodeXorr\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\txorl\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT( (Tree)b, "Invalid tree type in encodeXorr()\n");
 }
 return;
}

/* This function does encoding for the binary operator '|'.
 */
void encodeOrr(BinopTree b)
{
  TypeTag myType;

  if( !b )
    bugT( (Tree)b, "NULL pointer passed to encodeOrr\n");
  if( !(b->left && b->right))
    bugT( (Tree)b, "NULL child pointer passed to encodeOrr\n");
  
  b->expn.reg = setReg();
  myType = typeQuery(((ExpTree)b->left)->type);
  if( myType == TYERROR ) return; 
  switch ( myType ){
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%edx");
    emit("\tpopl\t%%eax");
    emit("\torl\t%%edx, %%eax");
    emit("\tpushl\t%%eax");
    return;
  }
  default:
    bugT((Tree)b, "Invalid tree type in encodeOrr()\n");
 }
  return;
}

/* Dummy function to encode leftAssign. It is never called.
 */
void encodeLeftAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeleftAssign: You are not supposed to get here!");
  return;
}

/* Dummy function to encode righttAssign. It is never called.
 */
void encodeRightAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeRightAssign: You are not supposed to get here!");
  return;
}

/* Dummy function to encode AndAssign. It is never called.
 */
void encodeAndAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeAndAssign: You are not supposed to get here!");
  return;
}

/* Dummy function to encode OrAssign. It is never called.
 */
void encodeOrAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeOrAssign: You are not supposed to get here!");
  return;
}

/* Dummy function to encode XorAssign. It is never called.
 */
void encodeXorAssign(BinopTree b)
{
  b = NULL; /* defeat warning */
  info("encodeXorAssign: You are not supposed to get here!");
  return;
}

/*  100% level  */

/* This function does the encoding for the postfix '++' operator.
 */
void encodeIncra(UnopTree u)
{
  /* error checkings*/
  if(!u)
    bugT((Tree)u, "Null pointer passed to encodeIncra\n");
  if(!(u->left) )
    bugT( (Tree)u, "Null child pointer passed to encodeIncra\n");

  /* getting the address */
  emit("\tmovl\t(%%esp), %%ecx");

  /* address popped and the value right on the stack*/
  encodeDerefDetail( u->expn.type, (Tree)u);
  
  /* addition */
  switch( typeQuery( u->expn.type) ){
  case TYFLOAT:
  {  
     emit("\tflds\t(%%esp)");    /* load the value */
     emit("\tfld1");             /* load 1.0 */
     emit("\tfaddp");            /* sum  and pop top f.p */
     emit("\tpushl\t%%ecx");     /* push the address on the stack */
     emit("\tsubl\t$4, %%esp");  /* reserve 4 bytes for the result*/
     emit("\tfstps\t(%%esp)");   /* put sum back to the run time stack */

     encodeAssignDetail( u->expn.type, (Tree)u); 
     emit("\taddl\t$4, %%esp");  /* pop off the value from the assignment */
     break;
  }
  
  case TYDOUBLE:
  {  emit("\tfldl\t(%%esp)");
     emit("\tpushl\t%%ecx");
     emit("\tfld1");
     emit("\tfaddp");
     emit("\tsubl\t$8, %%esp");  /* reserve 8 bytes for the result */
     emit("\tfstpl\t(%%esp)");

     encodeAssignDetail( u->expn.type, (Tree)u);
     emit("\taddl\t$8, %%esp");
     break;
  }
  
  case TYLONGDOUBLE:
  {  emit("\tfldt\t(%%esp)");
     emit("\tfld1");
     emit("\tfaddp");
     emit("\tpushl\t%%ecx");
     emit("\tsubl\t$12, %%esp");
     emit("\tfstpt\t(%%esp)");

     encodeAssignDetail( u->expn.type, (Tree)u );
     emit("\taddl\t$12, %%esp");
     break;
  }
  
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
     /* keep the value as returned value for postfix ++ */
     emit("\tmovl\t(%%esp), %%edx");
     emit("\tmovl\t$1, %%eax");
     emit("\taddl\t%%edx, %%eax");
     emit("\tpushl\t%%ecx"); /* push addr*/
     emit("\tpushl\t%%eax"); /* push result*/

     encodeAssignDetail( u->expn.type, (Tree)u );
     emit("\taddl\t$4, %%esp");
     break;
  }
  /* rel start */
  case TYPOINTER:
  {
    TypeQualifier qual;   /* temporary output paramater */
    int size;
    size = computeSize(typeQueryPointer(u->expn.type, &qual));
    emit("\tmovl\t(%%esp), %%edx");
    emit("\tmovl\t$%d, %%eax", size);
    emit("\taddl\t%%edx, %%eax");
    emit("\tpushl\t%%ecx"); /* push addr*/
    emit("\tpushl\t%%eax"); /* push result*/

    encodeAssignDetail( u->expn.type, (Tree)u );
    emit("\taddl\t$4, %%esp");
    break;
  }
  /* rel finish */
  default:
     bugT( (Tree)u, "invalid tree type in encodeIncra()\n");
 }
  u->expn.reg = setReg();
  return;
}

/* Dummy function for encoding prefix '++' operator.
 * It never gets called.
 */
void encodeIncrb(UnopTree u )
{
  u = NULL; /* defeat warning */
  info("encodeIncrb: You are not supposed to get here!\n");
  return; 
}

/* This function does the encoding for the prefix '++' operator.
 */
void encodeDecra(UnopTree u)
{
  /* error checkings */
  if( !u )
    bugT( (Tree)u, "Null pointer passed to encodeDecra\n");
  if( !(u->left))
    bugT( (Tree)u, "Null child pointer passed to encodeDecra\n");

  emit("\tmovl\t(%%esp), %%ecx");/* getting the address */
  /* address popped,the value right on the stack*/
  encodeDerefDetail( u->expn.type, (Tree)u);
  /* addition */
  switch( typeQuery( u->expn.type) ){
  case TYFLOAT:
  {  
     emit("\tflds\t(%%esp)"); /* load the value */
     emit("\tpushl\t%%ecx");
     emit("\tfld1"); /* load 1.0 */
     emit("\tfxch");
     emit("\tfsubp"); /* sum  and pop top f.p */
     emit("\tsubl\t$4, %%esp");/* reserve 4 bytes for the result*/
     emit("\tfstps\t(%%esp)");/* put sum back to the run time stack */

     encodeAssignDetail( u->expn.type, (Tree)u);
     emit("\taddl\t$4, %%esp");
     break;
  }
  case TYDOUBLE:
  {  emit("\tfldl\t(%%esp)");
     emit("\tpushl\t%%ecx");
     emit("\tfld1");
     emit("\tfxch");
     emit("\tfsubp");
     emit("\tsubl\t$8, %%esp");
     emit("\tfstpl\t(%%esp)");

     encodeAssignDetail( u->expn.type, (Tree)u);
     emit("\taddl\t$8, %%esp");
     break;
  }
  case TYLONGDOUBLE:
  {  emit("\tfldt\t(%%esp)");
     emit("\tfld1");
     emit("\tfxch");
     emit("\tfsubp");
     emit("\tpushl\t%%ecx");
     emit("\tsubl\t$12, %%esp");
     emit("\tfstpt\t(%%esp)");

     encodeAssignDetail( u->expn.type, (Tree)u );
     emit("\taddl\t$12, %%esp"); 
     break;
  }
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
     /* keep the value as returned value for postfix -- */
     emit("\tmovl\t(%%esp), %%edx");
     emit("\tmovl\t$1, %%eax");
     emit("\tsubl\t%%eax, %%edx");
     emit("\tpushl\t%%ecx"); /* push addr*/
     emit("\tpushl\t%%edx"); /* push result*/

     encodeAssignDetail( u->expn.type, (Tree)u );
     emit("\taddl\t$4, %%esp");
     break;
  }
  /* rel start */
  case TYPOINTER:
  {
    TypeQualifier qual;   /* temporary output paramater */
    int size;
    size = computeSize(typeQueryPointer(u->expn.type, &qual));
    emit("\tmovl\t(%%esp), %%edx");
    emit("\tmovl\t$%d, %%eax", size);
    emit("\tsubl\t%%eax, %%edx");
    emit("\tpushl\t%%ecx"); /* push addr*/
    emit("\tpushl\t%%edx"); /* push result*/

    encodeAssignDetail( u->expn.type, (Tree)u );
    emit("\taddl\t$4, %%esp");
    break;
  }
  /* rel finish */
  default:
     bugT( (Tree)u, "invalid tree type in encodeDecra()\n");
 }
  u->expn.reg = setReg();
  return;
}

/* Dummy function for encoding the prefix '++' operator.
 */
void encodeDecrb(UnopTree u )
{
  u = NULL; /* defeat warning */
  info("encodeDerb: You are not supposed to get here!\n");
  return;
}

/* Dummy function for encoding the cast operator.
 */
void encodeCast(CastopTree t )
{ 
  t = NULL; /* defeat warning */
  info("encodeCast: You are not supposed to get here!\n");
  return;
}

/*  This function encodes the comma operator.
 */
void encodeComma(BinopTree b)
{
  /* error checkings */
  if( !b )
    bugT( (Tree)b, "Null pointer passed to encodeComma\n");
  if( !b->left || !b->right )
    bugT( (Tree)b, "Null child pointer passed to encodeComma\n");

  /* evaluates the left operand first */
  encode( (Tree)(b->left));  
  
  /* pop the stack in order to have the left operand evaluated as void */
  freeExpressionStatementRegister( (Tree)b->left );
  
  /* evaluates the right operand, leaving the returned value on the stack */
  encode( (Tree)(b->right));

  /* mark the tree */
  b->expn.reg = setReg();
}

/* This function encodes the conditional operator.
 */
void encodeCond(TriopTree t)
{
  TypeTag myType;
  int lab1, lab2;

  lab1 = genLabel();
  lab2 = genLabel();

  /* error checkings */
  if( !t )
    bugT( (Tree)t, "Null pointer passed to encodeCond\n");
  if( !( t->left && t->middle && t->right ))
    bugT( (Tree)t, "Null child pointer passed to encodeCond\n");

  /* encode the first operand's exp. tree first, which has a Lnot operator */
  encode( (Tree)t->left );
  
  myType = typeQuery( ((ExpTree)t->left)->type );
  switch( myType ){
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  {
    emit("\tpopl\t%%eax");       
    emit("\tmovl\t$0, %%edx"); 
    emit("\tcmpl\t%%eax, %%edx"); 
    emit("\tjne\tL$%d", lab1 ); 
  }
  default:
    break;
 }
  /* encode the second operand's expression tree */
  encode( (Tree)t->middle ); 
  emit("\tjmp\tL$%d", lab2 );
  emit("L$%d:", lab1 );
  /* encode the third operand's expression tree */
  encode( (Tree)t->right ); 
  emit("L$%d:", lab2 );

  t->expn.reg = setReg();
  return;
}

/*  phase 5   */

/**
 * Encodes the AddrOf '&' operator. Relies on recursive calls from skeleton
 * to other methods to to encode the guts of it, so it actually does nothing. 
**/

void encodeAddrOf(UnopTree u)
{
  /* do nothing */
  u->expn.reg = setReg();
}

/**
 * Encodes the Index '[]' operator. Should never be called.
**/ 

void encodeIndex(BinopTree b)
{
  bugT((Tree)b, "encodeIndex should never be called");
}

/**
 * Encodes the '.' operator.
**/ 

void encodeFieldRef(FieldRefTree f)
{
   int offset;
   offset = computeMemberOffset(f->left->type, f->id);

   emit("\n#\t*************** start encode field ref **********");

   emit("\tpopl\t%%eax");              /* pop base address into eax */
   emit("\taddl\t$%d, %%eax", offset); /* add offset to base addr   */
   emit("\tpushl\t%%eax");             /* push new addr onto stack  */
   
   emit("#\t***************  end encode field ref  *********\n");
  
   f->expn.reg = setReg();
}

/**
 * Encodes the '->' operator. Should never be called.
**/

void encodePtr(PtrTree p)
{
  bugT((Tree)p, "encodePtr should never be called");
}
